import logging
from typing import Any, Iterable, List, Literal, Optional, Sequence, Set, Union

import boto3
from boto3.session import Session
from botocore.config import Config
from mypy_boto3_organizations.client import OrganizationsClient
from mypy_boto3_sts.client import STSClient

from botocove.cove_types import CoveSessionInformation

logger = logging.getLogger(__name__)


DEFAULT_ROLENAME = "OrganizationAccountAccessRole"


class CoveHostAccount(object):
    target_regions: Sequence[Optional[str]]

    def __init__(
        self,
        target_ids: Optional[List[str]],
        ignore_ids: Optional[List[str]],
        rolename: Optional[str],
        role_session_name: Optional[str],
        policy: Optional[str],
        policy_arns: Optional[List[str]],
        assuming_session: Optional[Session],
        org_master: bool,
        thread_workers: int,
        regions: Optional[List[str]],
    ) -> None:

        self.thread_workers = thread_workers

        self.sts_client = self._get_boto3_sts_client(assuming_session)
        self.org_client = self._get_boto3_org_client(assuming_session)

        self.host_account_id = self.sts_client.get_caller_identity()["Account"]

        if regions is None:
            self.target_regions = [None]
        else:
            self.target_regions = regions

        self.provided_ignore_ids = ignore_ids
        self.target_accounts = self._resolve_target_accounts(target_ids)
        if not self.target_accounts:
            raise ValueError(
                "There are no eligible account ids to run decorated func against"
            )

        self.role_to_assume = rolename or DEFAULT_ROLENAME
        self.role_session_name = role_session_name or self.role_to_assume
        self.policy = policy
        self.policy_arns = policy_arns

        self.org_master = org_master

    def get_cove_sessions(self) -> List[CoveSessionInformation]:
        logger.info(f"Getting session information for {self.target_accounts=}")
        logger.info(f"Role: {self.role_to_assume=} {self.role_session_name=}")
        logger.info(f"Session policy: {self.policy_arns=} {self.policy=}")
        return list(self._generate_account_sessions())

    def _generate_account_sessions(self) -> Iterable[CoveSessionInformation]:
        for region in self.target_regions:
            for account_id in self.target_accounts:
                account_details: CoveSessionInformation = CoveSessionInformation(
                    Id=account_id,
                    RoleName=self.role_to_assume,
                    RoleSessionName=self.role_session_name,
                    Policy=self.policy,
                    PolicyArns=self.policy_arns,
                    AssumeRoleSuccess=False,
                    Region=region,
                    ExceptionDetails=None,
                    Name=None,
                    Arn=None,
                    Email=None,
                    Status=None,
                    Result=None,
                )
                yield account_details

    def _get_boto3_client(
        self,
        clientname: Union[Literal["organizations"], Literal["sts"]],
        assuming_session: Optional[Session],
    ) -> Any:
        if assuming_session:
            logger.info(f"Using provided Boto3 session {assuming_session}")
            return assuming_session.client(
                service_name=clientname,
                config=Config(max_pool_connections=self.thread_workers),
            )
        logger.info("No Boto3 session argument: using credential chain")
        return boto3.client(
            service_name=clientname,
            config=Config(max_pool_connections=self.thread_workers),
        )

    def _get_boto3_org_client(
        self, assuming_session: Optional[Session]
    ) -> OrganizationsClient:
        client: OrganizationsClient = self._get_boto3_client(
            "organizations", assuming_session
        )
        return client

    def _get_boto3_sts_client(self, assuming_session: Optional[Session]) -> STSClient:
        client: STSClient = self._get_boto3_client("sts", assuming_session)
        return client

    def _resolve_target_accounts(self, target_ids: Optional[List[str]]) -> Set[str]:
        # Ensure we never run botocove on the account it's being run from
        validated_ignore_ids = {self.host_account_id}

        if self.provided_ignore_ids:
            validated_ignore_ids.update(self._parse_ignore_ids())
        logger.info(f"Ignoring account IDs: {validated_ignore_ids=}")

        if target_ids is None:
            # No target_ids passed, get all accounts in org
            target_accounts = self._get_active_org_accounts()
        else:
            # Specific list of IDs passed
            target_accounts = set(target_ids)

        return target_accounts - validated_ignore_ids

    def _parse_ignore_ids(self) -> Set[str]:
        if not isinstance(self.provided_ignore_ids, list):
            raise TypeError("ignore_ids must be a list of account IDs")
        for account_id in self.provided_ignore_ids:
            if len(account_id) != 12:
                raise TypeError("All ignore_id in list must be 12 character strings")
            if not isinstance(account_id, str):
                raise TypeError("All ignore_id list entries must be strings")
        return set(self.provided_ignore_ids)

    def _get_active_org_accounts(self) -> Set[str]:
        all_org_accounts = (
            self.org_client.get_paginator("list_accounts")
            .paginate()
            .build_full_result()["Accounts"]
        )
        return {acc["Id"] for acc in all_org_accounts if acc["Status"] == "ACTIVE"}



from typing import Any, Dict, List, Optional, TypedDict

from mypy_boto3_organizations.literals import AccountStatusType


class CoveSessionInformation(TypedDict):
    Id: str
    RoleName: str
    AssumeRoleSuccess: bool
    Arn: Optional[str]
    Email: Optional[str]
    Name: Optional[str]
    Status: Optional[AccountStatusType]
    RoleSessionName: Optional[str]
    Policy: Optional[str]
    PolicyArns: Optional[List[str]]
    Result: Any
    ExceptionDetails: Optional[Exception]
    Region: Optional[str]


CoveResults = List[CoveSessionInformation]


class CoveFunctionOutput(TypedDict):
    Results: CoveResults
    Exceptions: CoveResults


class CoveOutput(TypedDict):
    Results: List[Dict[str, Any]]
    Exceptions: List[Dict[str, Any]]
    FailedAssumeRole: List[Dict[str, Any]]



import logging
from concurrent import futures
from typing import Any, Callable

from tqdm import tqdm

from botocove.cove_host_account import CoveHostAccount
from botocove.cove_session import CoveSession
from botocove.cove_types import CoveFunctionOutput, CoveResults, CoveSessionInformation

logger = logging.getLogger(__name__)


class CoveRunner(object):
    def __init__(
        self,
        host_account: CoveHostAccount,
        func: Callable[..., Any],
        raise_exception: bool,
        func_args: Any,
        func_kwargs: Any,
        thread_workers: int,
    ) -> None:

        self.host_account = host_account
        self.sessions = host_account.get_cove_sessions()

        self.cove_wrapped_func = func
        self.raise_exception = raise_exception
        self.func_args = func_args
        self.func_kwargs = func_kwargs

        self.thread_workers = thread_workers

    def run_cove_function(self) -> CoveFunctionOutput:
        # Run decorated func with all valid sessions
        with futures.ThreadPoolExecutor(max_workers=self.thread_workers) as executor:
            completed: CoveResults = list(
                tqdm(
                    executor.map(self.cove_thread, self.sessions),
                    total=len(self.sessions),
                    desc="Executing function",
                    colour="#ff69b4",  # hotpink
                )
            )
        successful_results = [
            result for result in completed if not result["ExceptionDetails"]
        ]
        exceptions = [result for result in completed if result["ExceptionDetails"]]

        return CoveFunctionOutput(
            Results=successful_results,
            Exceptions=exceptions,
        )

    def cove_thread(
        self,
        account_session_info: CoveSessionInformation,
    ) -> CoveSessionInformation:
        cove_session = CoveSession(
            account_session_info,
            sts_client=self.host_account.sts_client,
            org_client=self.host_account.org_client,
            org_master=self.host_account.org_master,
        )
        try:
            cove_session.activate_cove_session()

            result = self.cove_wrapped_func(
                cove_session, *self.func_args, **self.func_kwargs
            )

            return cove_session.format_cove_result(result)

        except Exception as e:
            if self.raise_exception is True:
                logger.exception(cove_session.format_cove_error(e))
                raise
            else:
                return cove_session.format_cove_error(e)



import functools
import logging
from typing import Any, Callable, List, Optional

from boto3.session import Session

from botocove.cove_host_account import CoveHostAccount
from botocove.cove_runner import CoveRunner
from botocove.cove_types import CoveOutput

logger = logging.getLogger(__name__)


def cove(
    _func: Optional[Callable[..., Any]] = None,
    *,
    target_ids: Optional[List[str]] = None,
    ignore_ids: Optional[List[str]] = None,
    rolename: Optional[str] = None,
    role_session_name: Optional[str] = None,
    policy: Optional[str] = None,
    policy_arns: Optional[List[str]] = None,
    assuming_session: Optional[Session] = None,
    raise_exception: bool = False,
    org_master: bool = True,
    thread_workers: int = 20,
    regions: Optional[List[str]] = None
) -> Callable:
    def decorator(func: Callable[..., Any]) -> Callable[..., CoveOutput]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> CoveOutput:

            host_account = CoveHostAccount(
                target_ids=target_ids,
                ignore_ids=ignore_ids,
                rolename=rolename,
                role_session_name=role_session_name,
                policy=policy,
                policy_arns=policy_arns,
                org_master=org_master,
                assuming_session=assuming_session,
                thread_workers=thread_workers,
                regions=regions,
            )

            runner = CoveRunner(
                host_account=host_account,
                func=func,
                raise_exception=raise_exception,
                func_args=args,
                func_kwargs=kwargs,
                thread_workers=thread_workers,
            )

            output = runner.run_cove_function()

            # Rewrite dataclasses into untyped dicts to retain current functionality
            return CoveOutput(
                Results=[
                    {k: v for k, v in r.items() if v is not None}
                    for r in output["Results"]
                ],
                Exceptions=[
                    {k: v for k, v in e.items() if v is not None}
                    for e in output["Exceptions"]
                    if e["AssumeRoleSuccess"] is True
                ],
                FailedAssumeRole=[
                    {k: v for k, v in f.items() if v is not None}
                    for f in output["Exceptions"]
                    if f["AssumeRoleSuccess"] is False
                ],
            )

        return wrapper

    # Handle both bare decorator and with argument
    if _func is None:
        return decorator
    else:
        return decorator(_func)



import logging
from typing import Any

from boto3.session import Session
from botocore.exceptions import ClientError
from mypy_boto3_organizations.client import OrganizationsClient
from mypy_boto3_organizations.type_defs import AccountTypeDef
from mypy_boto3_sts.client import STSClient

from botocove.cove_types import CoveSessionInformation

logger = logging.getLogger(__name__)


class CoveSession(Session):
    """Enriches a boto3 Session with account data from Master account if run from
    an organization master.
    Provides internal helper functions for managing concurrent boto3 calls
    """

    assume_role_success: bool = False
    session_information: CoveSessionInformation
    stored_exception: Exception

    def __init__(
        self,
        session_info: CoveSessionInformation,
        org_client: OrganizationsClient,
        sts_client: STSClient,
        org_master: bool,
    ) -> None:
        self.session_information = session_info
        self.org_master = org_master
        self.org_client = org_client
        self.sts_client = sts_client

    def __repr__(self) -> str:
        # Overwrite boto3's repr to avoid AttributeErrors
        return f"{self.__class__.__name__}(account_id={self.session_information['Id']})"

    def activate_cove_session(self) -> "CoveSession":
        role_arn = (
            f"arn:aws:iam::{self.session_information['Id']}:role/"
            f"{self.session_information['RoleName']}"
        )

        if self.org_master:
            try:
                account_description: AccountTypeDef = self.org_client.describe_account(
                    AccountId=self.session_information["Id"]
                )["Account"]
                self.session_information["Arn"] = account_description["Arn"]
                self.session_information["Email"] = account_description["Email"]
                self.session_information["Name"] = account_description["Name"]
                self.session_information["Status"] = account_description["Status"]
            except ClientError:
                logger.exception(
                    "Failed to call describe_account for "
                    f"{self.session_information['Id']}"
                )

        try:
            logger.debug(f"Attempting to assume {role_arn}")

            # This calling style avoids a ParamValidationError from botocore.
            # Passing None is not allowed for the optional parameters.
            assume_role_args = {
                k: v
                for k, v in [
                    ("RoleArn", role_arn),
                    ("RoleSessionName", self.session_information["RoleSessionName"]),
                    ("Policy", self.session_information["Policy"]),
                    ("PolicyArns", self.session_information["PolicyArns"]),
                ]
                if v is not None
            }
            creds = self.sts_client.assume_role(**assume_role_args)["Credentials"]  # type: ignore[arg-type] # noqa E501

            init_session_args = {
                k: v
                for k, v in [
                    ("aws_access_key_id", creds["AccessKeyId"]),
                    ("aws_secret_access_key", creds["SecretAccessKey"]),
                    ("aws_session_token", creds["SessionToken"]),
                    ("region_name", self.session_information["Region"]),
                ]
                if v is not None
            }

            self.initialize_boto_session(**init_session_args)
            self.session_information["AssumeRoleSuccess"] = True
        except ClientError:
            logger.error(
                f"Failed to initalize cove session for "
                f"account {self.session_information['Id']}"
            )
            raise

        return self

    def initialize_boto_session(self, *args: Any, **kwargs: Any) -> None:
        # Inherit from and initialize standard boto3 Session object
        super().__init__(*args, **kwargs)

    def format_cove_result(self, result: Any) -> CoveSessionInformation:
        self.session_information["Result"] = result
        return self.session_information

    def format_cove_error(self, err: Exception) -> CoveSessionInformation:
        self.session_information["ExceptionDetails"] = err
        return self.session_information



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['botocove']

package_data = \
{'': ['*']}

install_requires = \
['boto3', 'boto3-stubs[sts,organizations]', 'tqdm']

setup_kwargs = {
    'name': 'botocove',
    'version': '1.5.2',
    'description': 'A decorator to allow running a function against all AWS accounts in an organization',
    'long_description': '# Botocove\n\nRun a function against a selection of AWS accounts, or all AWS accounts in an\norganization, concurrently with thread safety. Run in one or multiple regions.\n\nBy default, opinionated to work with the standard\nAWS Organization master/member configuration from an organization master\naccount.\n\n- Fast\n- Easy\n- Dolphin Themed 🐬\n\nA simple decorator for functions to remove time and complexity burden. Uses\n`ThreadPoolExecutor` to run boto3 sessions against AWS accounts concurrently.\n\nDecorating a function in `@cove` provides a boto3 session to the function and runs it\nin every account required, gathering all results into a dictionary.\n\n**Warning**: this tool gives you the potential to make dangerous changes\nat scale. **Test carefully and make idempotent changes**! Please read available\narguments to understand safe experimentation with this package.\n\n## Pre-requisites and Info\n\nAn IAM user with `sts:assumerole` privilege, and accounts that have a trust\nrelationship to the IAM user\'s account.\n\nBy default, the IAM user is expected to be in an AWS Organization Master account. You can alter nearly all behaviour of Cove with appropriate [arguments](#arguments)\n\nCove will not execute a function call in the account it\'s called from.\n\nDefault requirements are:\n\nIn the organization master account:\n* IAM permissions `sts:assumerole`, `sts:get-caller-identity` and\n`organizations:list-accounts`\n\nIn the organization accounts:\n* An `AccountOrganizationAccessRole` role\n\nSee the [arguments](#arguments) section for how to change these defaults to work with any\naccount configuration, including running without an AWS Organization.\n\n## Quickstart\nA function written to interact with one\n[boto3 session](https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html)\ncan now be called with a `session` from every account and region required by assuming a\n role in for each account - except the host account you\'re running from.\n\nFor example:\n\nA standard approach: this function takes a boto3 `session` and gets all IAM\nusers from a single AWS account. You would then manually run it in each account.\n\n```\nimport boto3\n\n\ndef get_iam_users(session):\n    iam = session.client("iam", region_name="eu-west-1")\n    all_users = iam.get_paginator("list_users").paginate().build_full_result()\n    return all_users\n\ndef main():\n    session = boto3.session.Session(profile_name="my_dev_account")\n    users = get_iam_users(session)\n    print(users) # A single account\'s IAM users\n```\n\nNow with `@cove`: a session for every account in the organization is injected\nby the decorator. A safe example to run as a test!\n\n```\nimport boto3\nfrom botocove import cove\n\n@cove()\ndef get_iam_users(session):\n    iam = session.client("iam", region_name="eu-west-1")\n    all_users = iam.get_paginator("list_users").paginate().build_full_result()\n    return all_users\n\ndef main():\n    # No session passed as the decorator injects it\n    all_results = get_iam_users()\n    # Now returns a Dict with keys Results, Exceptions and FailedAssumeRole\n\n    # A list of dictionaries for each account, with account details included.\n    # Each account\'s get_iam_users return is in a "Result" key.\n    print(all_results["Results"])\n\n    # A list of dictionaries for each account that raised an exception\n    print(all_results["Exceptions"])\n\n    # A list of dictionaries for each account that could not be assumed into\n    print(all_results["FailedAssumeRole"])\n```\n\nHere\'s an example of a more customised Cove decorator:\n```\n@cove(\n    target_ids=["123456789101", "234567891011"],\n    rolename="AWSControlTowerExecution",\n    raise_exception=True,\n    regions=["eu-west-1", "eu-west-2", "us-east-1"],\n)\ndef do_things(session):\n    # Cove will return six results of True, 2 accounts * 3 regions\n    return True\n```\n## Arguments\n\n### Cove\n`@cove()`:\n\nUses the standard boto3 credential chain to start with, assuming roles in every account\nrequired. Defaults to assuming the `OrganizationAccountAccessRole` in every account\nin an AWS organization.\n\nEquivalent to:\n`@cove(target_ids=None, ignore_ids=None, rolename=None, role_session_name=rolename,\n    policy=None, policy_arns=None, assuming_session=None, raise_exception=False,\n    org_master=True, thread_workers=20, regions=None)`\n\n`target_ids`: List[str]\n\nA list of AWS accounts as strings to attempt to assume role in to. When unset,\ndefault attempts to use every available account ID in an AWS organization.\n\n`ignore_ids`: List[str]\n\nA list of AWS account ID\'s that will not attempt assumption in to. Allows IDs to\nbe ignored.\n\n`rolename`: str\n\nAn IAM role name that will be attempted to assume in all target accounts.\nDefaults to the AWS Organization default, `OrganizationAccountAccessRole`.\n\n`role_session_name`: str\n\nAn IAM role session name that will be passed to each Cove session\'s `sts.assume_role()` call.\nDefaults to the name of the role being used if unset.\n\n`policy`: str\n\nA policy document that will be used as a session policy in each Cove session\'s `sts.assume_role()`\ncall. Unless the value is None, it is passed through via the Policy parameter.\n\n`policy_arns`: List[str]\n\nA list of managed policy ARNs that will be used as a session policy in each Cove session\'s\n `sts.assume_role()` call. Unless the value is None, it is passed through via the PolicyArns parameter.\n\n`assuming_session`: Session\n\nA Boto3 `boto3.session.Session()` object that will be used to call `sts:assumerole`. If not\nprovided, cove will instantiate one which will use the standard boto3 credential chain.\n\n`raise_exception`: bool\n\nDefaults to False. Default behaviour is that exceptions are not raised from\ndecorated function. This is due to `cove` running asynchronously and preferring\nto resolve all tasks and report their results instead of exiting early.\n\n`raise_exception=True` will allow a full stack trace to escape on the first\nexception seen; but will not gracefully or consistently interrupt running tasks.\nIt is vital to run interruptible, idempotent code with this argument as `True`.\n\n`org_master`: bool\n\nDefaults to True. When True, will leverage the Boto3 Organizations API to list\nall accounts in the organization, and enrich each `CoveSession` with information\navailable (`Id`, `Arn`, `Name`, `Status`, `Email`). Disabling this and providing your\nown full list of accounts may be a desirable optimisation if speed is an issue.\n\n`org_master=False` means only `Id` will be\navailable to `CoveSession`.\n\n`thread_workers`: int\n\nDefaults to 20. Cove utilises a ThreadPoolWorker under the hood, which can be tuned\nwith this argument. Number of thread workers directly corrolates to memory usage: see\n[here](#is-botocove-thread-safe)\n\n`regions`: List[str]\n\nIf not provided, Cove will respect your profile\'s default region via the boto credential chain.\nIf provided, Cove will run the decorated function in every region named.\n\nYou can get all regions with:\n`regions = [r[\'RegionName\'] for r in boto3.client(\'ec2\').describe_regions()[\'Regions\']]`✨\n\n### CoveSession\n\nCove supplies an enriched Boto3 session to each function called. Account details\nare available with the `session_information` attribute.\n\nOtherwise, it functions exactly as calling `boto3` would.\n\n```\n@cove()\ndef do_nothing(session: CoveSession):\n    print(session.session_information) # Outputs a dict of known information\n    # This function runs no boto3-specific API, but does test that a role\n    # can be assumed\n```\n\n## Return values\n\nWrapped functions return a dictionary. Each value contains List[Dict[str, Any]]:\n```\n{\n    "Results": results:\n    "Exceptions": exceptions,\n    "FailedAssumeRole": invalid_sessions,\n}\n```\nAn example of cove_output["Results"]:\n```\n[ # A list of dictionaries per account called\n    {\n    \'Id\': \'123456789010\',\n    \'Email\': \'email@address.com\',\n    \'Name\': \'account-name\',\n    \'Status\': \'ACTIVE\',\n    \'AssumeRoleSuccess\': True,\n    \'Result\': wrapped_function_return_value # Result of wrapped func\n    }\n]\n```\n\n### Is botocove thread safe?\n\nbotocove is thread safe, but number of threaded executions will be bound by memory,\nnetwork IO and AWS api rate limiting. Defaulting to 20 thread workers is a reasonable\nstarting point, but can be further optimised for runtime with experimentation.\n\nbotocove has no constraint or understanding of the function it\'s wrapping: it is\nrecommended to avoid shared state for botocove wrapped functions, and to write simple\nfunctions that are written to be idempotent and independent.\n\n[Boto3 Session objects are not natively thread safe and should not be shared across threads](https://boto3.amazonaws.com/v1/documentation/api/1.14.31/guide/session.html#multithreading-or-multiprocessing-with-sessions).\nHowever, botocove is instantiating a new Session object per thread/account and running\ndecorated functions inside their own closure. A shared client is used from the host account\nthat botocove is run from (eg, an organization master account) -\n[clients are threadsafe](https://boto3.amazonaws.com/v1/documentation/api/latest/guide/clients.html#multithreading-or-multiprocessing-with-clients) and allow this.\n\nboto3 sessions have a significant memory footprint:\nVersion 1.5.0 of botocove was re-written to ensure that boto3 sessions are released\nafter completion which resolved memory starvation issues. This was discussed here:\nhttps://github.com/connelldave/botocove/issues/20 and a relevant boto3 issue is here:\nhttps://github.com/boto/boto3/issues/1670\n\n### botocove?\n\nIt turns out that the Amazon\'s Boto dolphins are solitary or small-group animals,\nunlike the large pods of dolphins in the oceans. This killed my "large group of\nboto" idea, so the next best idea was where might they all shelter together... a\ncove!\n',
    'author': 'Dave Connell',
    'author_email': 'daveconn41@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/connelldave/botocove',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
